/*
 * emdadm - manage Linux "emd" devices aka RAID arrays.
 *
 * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
 *
 * Added pluggable metadata support.
 * Copyright (C) 2004 Adaptec Inc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *    Author: Neil Brown
 *    Email: <neilb@cse.unsw.edu.au>
 *    Paper: Neil Brown
 *           School of Computer Science and Engineering
 *           The University of New South Wales
 *           Sydney, 2052
 *           Australia
 */

#include "mdadm.h"
#include "md_u.h"
#include "md_p.h"

#define REGISTER_DEV 		_IO (MD_MAJOR, 1)
#define START_MD     		_IO (MD_MAJOR, 2)
#define STOP_MD      		_IO (MD_MAJOR, 3)

int Manage_ro(char *devname, int fd, int readonly)
{
	/* switch to readonly or rw
	 *
	 * first check that array is runing
	 * use RESTART_ARRAY_RW or STOP_ARRAY_RO
	 *
	 */
	mdu_array_info_t array;
		
	if (ioctl(fd, GET_ARRAY_INFO, &array)) {
		fprintf(stderr, Name ": %s does not appear to be active.\n",
			devname);
		return 1;
	}
		
	if (readonly>0) {
		if (ioctl(fd, EMD_STOP_ARRAY_RO, array.cookie)) {
			fprintf(stderr, Name ": failed to set readonly for %s: %s\n",
				devname, strerror(errno));
			return 1;
		}
	} else if (readonly < 0) {
		if (ioctl(fd, EMD_RESTART_ARRAY_RW, array.cookie)) {
			fprintf(stderr, Name ": failed to set writable for %s: %s\n",
				devname, strerror(errno));
			return 1;
		}
	}
	return 0;			
}

int Manage_runstop(char *devname, int fd, int runstop)
{
	unsigned long cookie;	
	mdu_array_info_t array;
        	
	cookie = get_cookie(devname); 
	if (cookie == 0) {
		fprintf(stderr, Name " cannot find device %s\n", devname );
		return 1;
	}
	
	if (runstop>0) {
		if (ioctl(mdctl_fd, EMD_RUN_ARRAY, cookie)) {
		        fprintf(stderr, Name ": failed to run array %s: %s\n",
				devname, strerror(errno));
			return 1;
		}
	} else if (runstop < 0){
		if (get_array_info(devname, &array) != 0) {
			fprintf(stderr, Name ": %s does not appear to be active.\n",
				devname);
			return 1;
		}
		if (ioctl(mdctl_fd, EMD_STOP_ARRAY, cookie)) {
			fprintf(stderr, Name ": fail to stop array %s: %s\n",
				devname, strerror(errno));
			return 1;
		}
	}
	return 0;
}

int Manage_subdevs(char *devname, int fd, mddev_dev_t devlist)
{
	/* do something to each dev.
	 * devmode can be
	 *  'a' - add the device
	 *	   try HOT_ADD_DISK
	 *         If that fails EINVAL, try ADD_NEW_DISK
	 *  'r' - remove the device HOT_REMOVE_DISK
	 *  'f' - set the device faulty SET_DISK_FAULTY
	 */
	mdu_array_info_t array;
	mdu_disk_info_t disc;
        dev_t device;
        mdu_major_minor_t major_minor; 
	mddev_dev_t dv;
	struct stat stb;
	int j;
	int save_errno;
	static char buf[4096];

	memset(&disc, 0, sizeof(disc));
	
	if (get_array_info(devname, &array)) {
		fprintf(stderr, Name ": cannot get array info for %s\n",
			devname);
		return 1;
	}
	
	for (dv = devlist ; dv; dv=dv->next) {
		if (stat(dv->devname, &stb)) {
			fprintf(stderr, Name ": cannot find %s: %s\n",
				dv->devname, strerror(errno));
			return 1;
		}
		if ((stb.st_mode & S_IFMT) != S_IFBLK) {
			fprintf(stderr, Name ": %s is not a block device.\n",
				dv->devname);
			return 1;
		}
		
		device = get_devnum(dv->devname); 
	      	major_minor.major = major(device);
      		major_minor.minor = minor(device);

		switch(dv->disposition){
		default:
			fprintf(stderr, Name ": internal error - devmode[%s]=%d\n",
				dv->devname, dv->disposition);
			return 1;
		case 'a':
			save_errno = errno;
			if (read(fd, buf, sizeof(buf)) > 0) {
				/* array is active, so don't try to add.
				 * i.e. something is wrong 
				 */
				fprintf(stderr, Name ": hot add failed for %s: %s\n",
					dv->devname, strerror(save_errno));
				return 1;
			}
			/* try ADD_NEW_DISK.
			 * set up number/raid_disk/state just
			 * in case
			 */
			for (j = 0; j < array.nr_disks; j++) {
				disc.number = j;
				if (get_disk_info(dv->devname, &disc))
					break;
				if (disc.major==0 && disc.minor==0)
					break;
				if (disc.state & 8) /* removed */
					break;
			}
			disc.number = j;
			disc.raid_disk = j;
			disc.state = 0;
			disc.major = MAJOR(stb.st_rdev);
			disc.minor = MINOR(stb.st_rdev);
			
			// TBD: Probably don't need this
			disc.cookie = get_cookie(devname);
			disc.parent_cookie = get_cookie(devname);

			if (ioctl(mdctl_fd, EMD_HOT_ADD_DISK, &disc)) {
				fprintf(stderr, Name ": add new device failed for %s: %s\n",
					dv->devname, strerror(errno));
				return 1;
			}
			fprintf(stderr, Name ": added %s\n", dv->devname);
			break;

		case 'r': /* Hot Remove */
			disc.major = MAJOR(stb.st_rdev);
			disc.minor = MINOR(stb.st_rdev);
			disc.parent_cookie = get_cookie(devname);
			if (ioctl(mdctl_fd, EMD_HOT_REMOVE_DISK, &disc)) {
				fprintf(stderr, Name ": hot remove failed for %s: %s\n",
					dv->devname, strerror(errno));
				return 1;
			}
			fprintf(stderr, Name ": hot removed %s\n", dv->devname);
			break;

		case 'f': /* Set Faulty */
			disc.major = MAJOR(stb.st_rdev);
			disc.minor = MINOR(stb.st_rdev);
			disc.cookie = get_cookie(devname);
			//disc.parent_cookie = get_cookie(devname);
			if (ioctl(mdctl_fd, EMD_SET_DISK_FAULTY, &disc)) {
				fprintf(stderr, Name ": set device faulty failed for %s:  %s\n",
					dv->devname, strerror(errno));
				return 1;
			}
			fprintf(stderr, Name ": set %s faulty in %s\n",
				dv->devname, devname);
			break;
		}
	}
	return 0;
	
}

int Manage_delete(char *devname, int fd, int runstop)
{
	unsigned long cookie;	
	mdu_array_info_t array;
	mdu_disk_info_t *disk;
	int member_cnt;
	int i;
	char *dv;
        	
	cookie = get_cookie(devname); 
	if (cookie == 0) {
		fprintf(stderr, Name " cannot find device %s\n", devname );
		return 1;
	}
	
	if (runstop>0) {
		if (ioctl(mdctl_fd, EMD_RUN_ARRAY, cookie)) {
		        fprintf(stderr, Name ": failed to run array %s: %s\n",
				devname, strerror(errno));
			return 1;
		}
	} else if (runstop < 0){
		if (get_array_info(devname, &array) != 0) {
			fprintf(stderr, Name ": %s does not appear to be active.\n",
				devname);
			return 1;
		}
		disk = get_array_member_info(&array, devname);
		member_cnt = array.raid_disks + array.spare_disks;
		if (ioctl(mdctl_fd, EMD_STOP_ARRAY, cookie)) {
			fprintf(stderr, Name ": fail to stop array %s: %s\n",
				devname, strerror(errno));
			return 1;
		}
		if(disk != NULL){
			for(i=0; i<member_cnt; i++){
				if ((dv = map_dev(disk[i].major, 
					disk[i].minor))) {
					Kill(dv, /*force kill*/1);
				}
			}
		}
	}
	return 0;
}

mdu_disk_info_t *get_array_member_info(mdu_array_info_t *array, char *array_name)
{
        int d, rc, member_cnt;
	unsigned long parent_cookie;
	mdu_disk_info_t *disk;
	mdu_disk_info_t *pdisk;
	
        parent_cookie = array->cookie;
	member_cnt = array->raid_disks + array->spare_disks;

	if((disk = malloc(sizeof(mdu_disk_info_t)*member_cnt)) == NULL)
		return NULL;

	pdisk = disk;

	for (d = 0; d < member_cnt; d++) {
                rc = get_disk_info_indexed(parent_cookie, d, &disk[d]);
		if ( rc != 0) {
			if (d < array->raid_disks) 
				disk[d].state = MDU_MSTATE_FAILED;
			else
				continue;
		}
	}
	return pdisk;
}
