/*
 * emdadm - manage Linux "emd" devices aka RAID arrays.
 *
 * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
 *
 * Added pluggable metadata support.
 * Copyright (C) 2004 Adaptec Inc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *    Author: Neil Brown
 *    Email: <neilb@cse.unsw.edu.au>
 *    Paper: Neil Brown
 *           School of Computer Science and Engineering
 *           The University of New South Wales
 *           Sydney, 2052
 *           Australia
 */

#include	"mdadm.h"
#include	"md_p.h"
#include	<sys/utsname.h>
#include	<ctype.h>

/*
 * Parse a 128 bit uuid in 4 integers
 * format is 32 hexx nibbles with options :.<space> separator
 * If not exactly 32 hex digits are found, return 0
 * else return 1
 */
int parse_uuid(char *str, int uuid[4])
{
    int hit = 0; /* number of Hex digIT */
    int i;
    char c;
    for (i=0; i<4; i++) uuid[i]=0;

    while ((c= *str++)) {
	int n;
	if (c>='0' && c<='9')
	    n = c-'0';
	else if (c>='a' && c <= 'f')
	    n = 10 + c - 'a';
	else if (c>='A' && c <= 'F')
	    n = 10 + c - 'A';
	else if (strchr(":. -", c))
	    continue;
	else return 0;

	if (hit<32) {
	    uuid[hit/8] <<= 4;
	    uuid[hit/8] += n;
	}
	hit++;
    }
    if (hit == 32)
	return 1;
    return 0;
}


/*
 * Get the md version number.
 * We use the RAID_VERSION ioctl if it is supported
 * If not, but we have a block device with major '9', we assume
 * 0.36.0
 *
 * Return version number as 24 but number - assume version parts
 * always < 255
 */

int md_get_version()
{
    struct stat stb;
    mdu_version_t vers;
    int fd = open_mdctl();

    if ((fd < 0) || (fstat(fd, &stb) < 0) || 
	((S_IFMT & stb.st_mode) != S_IFCHR)){
	close_mdctl();
	return -1;	
    }

    if (ioctl(fd, RAID_VERSION, &vers) == 0){
    	close_mdctl();
	return  (vers.major*10000) + (vers.minor*100) + vers.patchlevel;
    }

    if (errno == EACCES){
	close_mdctl();	
	return -1;
    }

    if (MAJOR(stb.st_rdev) == MD_MAJOR){
	close_mdctl();
	return (3600);
    }
    close_mdctl();
    return -1;
}

    
int get_linux_version()
{
	struct utsname name;
	char *cp;
	int a,b,c;
	if (uname(&name) <0)
		return -1;

	cp = name.release;
	a = strtoul(cp, &cp, 10);
	if (*cp != '.') return -1;
	b = strtoul(cp+1, &cp, 10);
	if (*cp != '.') return -1;
	c = strtoul(cp+1, NULL, 10);

	return (a*1000000)+(b*1000)+c;
}

int enough(int level, int raid_disks, int avail_disks)
{
	switch (level) {
	case -4:
		return avail_disks>= 1;
	case -1:
	case 0:
		return avail_disks == raid_disks;
	case 1:
		return avail_disks >= 1;
	case 4:
	case 5:
		return avail_disks >= raid_disks-1;
	case 6:
		return avail_disks >= raid_disks-2;
	default:
		return 0;
	}
}

int same_uuid(int a[4], int b[4])
{
    if (a[0]==b[0] &&
	a[1]==b[1] &&
	a[2]==b[2] &&
	a[3]==b[3])
	return 1;
    return 0;
}

void uuid_from_super(int uuid[4], mdp_super_t *super)
{
    uuid[0] = super->set_uuid0;
    if (super->minor_version >= 90) {
	uuid[1] = super->set_uuid1;
	uuid[2] = super->set_uuid2;
	uuid[3] = super->set_uuid3;
    } else {
	uuid[1] = 0;
	uuid[2] = 0;
	uuid[3] = 0;
    }
}

int compare_super(mdp_super_t *first, mdp_super_t *second)
{
    /*
     * return:
     *  0 same, or first was empty, and second was copied
     *  1 second had wrong number
     *  2 wrong uuid
     *  3 wrong other info
     */
    int uuid1[4], uuid2[4];
    if (second->md_magic != MD_SB_MAGIC)
	return 1;
    if (first-> md_magic != MD_SB_MAGIC) {
	memcpy(first, second, sizeof(*first));
	return 0;
    }

    uuid_from_super(uuid1, first);
    uuid_from_super(uuid2, second);
    if (!same_uuid(uuid1, uuid2))
	return 2;
    if (first->major_version != second->major_version ||
	first->minor_version != second->minor_version ||
	first->patch_version != second->patch_version ||
	first->gvalid_words  != second->gvalid_words  ||
	first->ctime         != second->ctime         ||
	first->level         != second->level         ||
	first->size          != second->size          ||
	first->raid_disks    != second->raid_disks    )
	return 3;

    return 0;
}

int load_super(int fd, mdp_super_t *super)
{
	/* try to read in the superblock
	 *
	 * return
	 *   0 - success
	 *   1 - no block size
	 *   2 - too small
	 *   3 - no seek
	 *   4 - no read
	 *   5 - no magic
	 *   6 - wrong major version
	 */
	unsigned long size;
	unsigned long long offset;
    
	if (ioctl(fd, BLKGETSIZE, &size))
		return 1;

	if (size < MD_RESERVED_SECTORS*2)
		return 2;
	
	offset = MD_NEW_SIZE_SECTORS(size);

	offset *= 512;

	ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */

	if (lseek64(fd, offset, 0)< 0LL)
		return 3;

	if (read(fd, super, sizeof(*super)) != sizeof(*super))
		return 4;

	if (super->md_magic != MD_SB_MAGIC)
		return 5;

	if (super->major_version != 0)
		return 6;
	return 0;
}

int store_super(int fd, mdp_super_t *super)
{
	long size;
	long long offset;
    
	if (ioctl(fd, BLKGETSIZE, &size))
		return 1;

	if (size < MD_RESERVED_SECTORS*2)
		return 2;
	
	offset = MD_NEW_SIZE_SECTORS(size);

	offset *= 512;

	if (lseek64(fd, offset, 0)< 0LL)
		return 3;

	if (write(fd, super, sizeof(*super)) != sizeof(*super))
		return 4;

	return 0;
}
    

int check_ext2(int fd, char *name)
{
	/*
	 * Check for an ext2fs file system.
	 * Superblock is always 1K at 1K offset
	 *
	 * s_magic is le16 at 56 == 0xEF53
	 * report mtime - le32 at 44
	 * blocks - le32 at 4
	 * logblksize - le32 at 24
	 */
	unsigned char sb[1024];
	time_t mtime;
	int size, bsize;
	if (lseek(fd, 1024,0)!= 1024)
		return 0;
	if (read(fd, sb, 1024)!= 1024)
		return 0;
	if (sb[56] != 0x53 || sb[57] != 0xef)
		return 0;

	mtime = sb[44]|(sb[45]|(sb[46]|sb[47]<<8)<<8)<<8;
	bsize = sb[24]|(sb[25]|(sb[26]|sb[27]<<8)<<8)<<8;
	size = sb[4]|(sb[5]|(sb[6]|sb[7]<<8)<<8)<<8;
	fprintf(stderr, Name ": %s appears to contain an ext2fs file system\n",
		name);
	fprintf(stderr,"    size=%dK  mtime=%s",
		size*(1<<bsize), ctime(&mtime));
	return 1;
}


int check_reiser(int fd, char *name)
{
	/*
	 * superblock is at 64K
	 * size is 1024;
	 * Magic string "ReIsErFs" or "ReIsEr2Fs" at 52
	 *
	 */
	unsigned char sb[1024];
	int size;
	if (lseek(fd, 64*1024, 0) != 64*1024)
		return 0;
	if (read(fd, sb, 1024) != 1024)
		return 0;
	if (strncmp(sb+52, "ReIsErFs",8)!=0 &&
	    strncmp(sb+52, "ReIsEr2Fs",9)!=0)
		return 0;
	fprintf(stderr, Name ": %s appears to contain a reiserfs file system\n",name);
	size = sb[0]|(sb[1]|(sb[2]|sb[3]<<8)<<8)<<8;
	fprintf(stderr, "    size = %dK\n", size*4);
		
	return 1;
}


int check_raid(int fd, char *name)
{
	time_t crtime;
        int rc;
        mdu_array_info_t array;

        rc = get_array_info( name, &array );
	if (rc != 0)
		return 0;

	/* Looks like a raid array .. */
	fprintf(stderr, Name ": %s appears to be part of a raid array:\n",
		name);
	crtime = array.ctime;
	fprintf(stderr, "    level=%d devices=%d ctime=%s",
		array.level, array.raid_disks, ctime(&crtime));
	return 1;
}


int ask(char *mesg)
{
	char *add = "";
	int i;
	return 1;
	for (i=0; i<5; i++) {
		char buf[100];
		fprintf(stderr, "%s%s", mesg, add);
		fflush(stderr);
		fflush(stdin);
		if (fgets(buf, 100, stdin)==NULL)
			return 0;
		if (buf[0]=='y' || buf[0]=='Y')
			return 1;
		if (buf[0]=='n' || buf[0]=='N')
			return 0;
		add = "(y/n) ";
	}
	fprintf(stderr, Name ": assuming 'no'\n");
	return 0;
}


char *map_num(mapping_t *map, int num)
{
	while (map->name) {
		if (map->num == num)
			return map->name;
		map++;
	}
	return NULL;
}


int map_name(mapping_t *map, char *name)
{
	while (map->name) {
		if (strcmp(map->name, name)==0)
			return map->num;
		map++;
	}
	return UnSet;
}


/*
 * convert a major/minor pair for a block device into a name in /dev, if possible.
 * On the first call, walk /dev collecting name.
 * Put them in a simple linked listfor now.
 */
struct devmap {
    int major, minor;
    char *name;
    struct devmap *next;
} *devlist = NULL;
int devlist_ready = 0;

#ifdef UCLIBC
char *map_dev(int major, int minor)
{
#if 0
	fprintf(stderr, "Warning - fail to map %d,%d to a device name\n",
		major, minor);
#endif
	return NULL;
}
#else
#define  __USE_XOPEN_EXTENDED
#include <ftw.h>


#ifndef __dietlibc__
int add_dev(const char *name, const struct stat *stb, int flag, struct FTW *s)
#else
int add_dev(const char *name, const struct stat *stb, int flag)
#endif
{
    if ((stb->st_mode&S_IFMT)== S_IFBLK) {
	char *n = strdup(name);
	struct devmap *dm = malloc(sizeof(*dm));
	if (dm) {
	    dm->major = MAJOR(stb->st_rdev);
	    dm->minor = MINOR(stb->st_rdev);
	    dm->name = n;
	    dm->next = devlist;
	    devlist = dm;
	}
    }
    return 0;
}


char *map_dev(int major, int minor)
{
    struct devmap *p;
    if (!devlist_ready) {
#ifndef __dietlibc__
	nftw("/dev", add_dev, 10, FTW_PHYS);
#else
	ftw("/dev", add_dev, 10);
#endif
	devlist_ready=1;
    }

    for (p=devlist; p; p=p->next)
	if (p->major == major &&
	    p->minor == minor)
	    return p->name;
    return NULL;
}

#endif

unsigned long calc_sb_csum(mdp_super_t *super)
{
        unsigned int  oldcsum = super->sb_csum;
	unsigned long long newcsum = 0;
	unsigned long csum;
	int i;
	unsigned int *superc = (int*) super;
	super->sb_csum = 0;

	for(i=0; i<MD_SB_BYTES/4; i++)
		newcsum+= superc[i];
	csum = (newcsum& 0xffffffff) + (newcsum>>32);
	super->sb_csum = oldcsum;
	return csum;
}


char *human_size(long long bytes)
{
	static char buf[30];
	

	if (bytes < 5000*1024)
		buf[0]=0;
	else if (bytes < 2*1024LL*1024LL*1024LL)
		sprintf(buf, " %ld.%02ld MiB %ld.%02ld MB",
			(long)(bytes>>20),
			(long)((bytes&0xfffff)+0x100000/200)/(0x100000/100),
			(long)(bytes/1000/1000),
			(long)(((bytes%1000000)+5000)/10000)
			);
	else
		sprintf(buf, " %ld.%02ld GiB %ld.%02ld GB",
			(long)(bytes>>30),
			(long)(((bytes>>10)&0xfffff)+0x100000/200)/(0x100000/100),
			(long)(bytes/1000LL/1000LL/1000LL),
			(long)((((bytes/1000)%1000000)+5000)/10000)
			);
	return buf;
}


char *mb_size(long long bytes)
{
	static char buf[30];
	

	if (bytes < 5000*1024)
		buf[0]=0;
	else if (bytes < 2*1024LL*1024LL*1024LL)
		sprintf(buf, " %ld.%02ld MB",
			(long)(bytes/1000/1000),
			(long)(((bytes%1000000)+5000)/10000)
			);
	else
		sprintf(buf, " %ld.%02ld GB",
			(long)(bytes/1000LL/1000LL/1000LL),
			(long)((((bytes/1000)%1000000)+5000)/10000)
			);
	return buf;
}


char *human_size_brief(long long bytes)
{
	static char buf[30];
	

	if (bytes < 5000*1024)
		sprintf(buf, "%ld.%02ldKiB",
			(long)(bytes>>10), (long)(((bytes&1023)*100+512)/1024)
			);
	else if (bytes < 2*1024LL*1024LL*1024LL)
		sprintf(buf, "%ld.%02ldMiB",
			(long)(bytes>>20),
			(long)((bytes&0xfffff)+0x100000/200)/(0x100000/100)
			);
	else
		sprintf(buf, "%ld.%02ldGiB",
			(long)(bytes>>30),
			(long)(((bytes>>10)&0xfffff)+0x100000/200)/(0x100000/100)
			);
	return buf;
}


static int mdp_major = -1;
void get_mdp_major(void)
{
	FILE *fl = fopen("/proc/devices", "r");
	char *w;
	int have_block = 0;
	int have_devices = 0;
	int last_num = -1;
	if (!fl)
		return;
	while ((w = conf_word(fl, 1))) {
		if (have_block && strcmp(w, "devices:")==0)
			have_devices = 1;
		have_block =  (strcmp(w, "Block")==0);
		if (isdigit(w[0]))
			last_num = atoi(w);
		if (have_devices && strcmp(w, "mdp")==0)
			mdp_major = last_num;
		free(w);
	}
	fclose(fl);
}



char *get_md_name(int dev)
{
	/* find /dev/emd%d or /dev/emd/%d or make a device /dev/.tmp.md%d */
	/* if dev < 0, want /dev/emd/d%d or find mdp in /proc/devices ... */
	static char devname[50];
	struct stat stb;
	dev_t rdev;

	if (dev < 0) {

		if (mdp_major < 0) get_mdp_major();
		if (mdp_major < 0) return NULL;
		rdev = MKDEV(mdp_major, (-1-dev)<<6);
		sprintf(devname, "/dev/emd/d%d", -1-dev);
		if (stat(devname, &stb) == 0
		    && (S_IFMT&stb.st_mode) == S_IFBLK
		    && (stb.st_rdev == rdev))
			return devname;
	} else {
		rdev = MKDEV(MDP_MAJOR, dev << MDP_MINOR_SHIFT);
		/* convert index to ascii. 0 -> 'a', 1 -> 'b' */
		if (LINUX_VERSION < KERNEL_VERSION(2,6,0))
			sprintf(devname, "/dev/emd%c", dev + 'a'); 
		else
			sprintf(devname, "/dev/emd/%c", dev + '0'); 
		if (stat(devname, &stb) == 0
		    && (S_IFMT&stb.st_mode) == S_IFBLK
		    && (stb.st_rdev == rdev))
			return devname;

		sprintf(devname, "/dev/emd/%d", dev);
		if (stat(devname, &stb) == 0
		    && (S_IFMT&stb.st_mode) == S_IFBLK
		    && (stb.st_rdev == rdev))
			return devname;
	}
	sprintf(devname, "/dev/.tmp.emd%d", dev);
	if (mknod(devname, S_IFBLK | 0600, rdev) == -1)
		return NULL;

	if (stat(devname, &stb) == 0
	    && (S_IFMT&stb.st_mode) == S_IFBLK
	    && (stb.st_rdev == rdev))
		return devname;
	unlink(devname);
	return NULL;
}


void put_md_name(char *name)
{
	if (strncmp(name, "/dev/.tmp.md", 12)==0)
		unlink(name);
}


/****************************************************/
/* open_mdctl()   Open the EMD IOCTL Control Device */
/****************************************************/

int open_mdctl()
{
	mdctl_fd = open("/dev/emdctl", O_RDWR, 0);

	if (mdctl_fd < 0)
		fprintf(stderr, Name ": error opening /dev/emdctl: %s\n", strerror(errno));
	
	return mdctl_fd;
}


/*****************************************************/
/* close_mdctl()  Close the EMD IOCTL Control Device */
/*****************************************************/

void close_mdctl()
{
	if (mdctl_fd > 0)
		close(mdctl_fd);   
 	return;
}


/**********************************************************/
/* get_devnum()   Return a dev_t number for a device name */
/**********************************************************/

dev_t get_devnum(char *devname)
{
        struct stat sbuf;  

        if  ( stat( devname, &sbuf ) )
             printf( "Stat( %s Failed\n", devname );
         
        return( sbuf.st_rdev );
}


/***************************************************************/
/* get_array_info()   Get the EMD Array Info for a device name */
/***************************************************************/

int get_array_info(char *device_name, mdu_array_info_t *array_info)
{
        int rc;

        /* First, zero out the entire array structure */
        memset( ( void * ) array_info, 0, sizeof( mdu_array_info_t ) ); 

        /* Set the cookie value in the Array structure */
        array_info->cookie = get_cookie( device_name );
        	 
        /* Now, Ask the EMD driver to give us the array info */
	rc = ioctl(mdctl_fd, EMD_GET_ARRAY_INFO, array_info);
//      if  ( rc )    
//             printf( "Error in EMD_GET_ARRAY_INFO Ioctl = %X : %s\n",
//	       		rc,  strerror(errno));

	return rc;
}


/*********************************************************/
/* get_disk_info()   Get EMD Disk Info for a device name */
/*********************************************************/

int get_disk_info(char *device_name, mdu_disk_info_t *disk_info)
{
        int rc;
        int fd; 
	struct stat stb;
	
        /* Zero out the entire Disk structure */
        memset( ( void * ) disk_info, 0, sizeof( mdu_disk_info_t ) ); 

        /* Set the Cookie value in the Disk structure */
        disk_info->cookie = get_cookie( device_name );

	/* Set the Device Numbers in the Disk Structure */	
	fd = open(device_name, O_RDONLY, 0); 
	fstat(fd, &stb); 
	disk_info->major = MAJOR(stb.st_rdev);	
	disk_info->minor = MINOR(stb.st_rdev);
	close(fd);
        
	/* Now, Ask the EMD driver to give us the disk info */
	rc = ioctl(mdctl_fd, EMD_GET_DISK_INFO, disk_info);
//      if  ( rc )   
//            printf( "Error in EMD_GET_DISK_INFO Ioctl = %X : %s\n",
//	 	       rc,  strerror(errno));

	return rc;
}


/*****************************************************************************/
/* get_disk_info_indexed()   Get EMD Disk Info device using Parent and Index */
/*****************************************************************************/

int get_disk_info_indexed(unsigned long parent_cookie, int index, mdu_disk_info_t *disk_info)
{
        int rc;

        /* Zero out the entire Disk structure */
        memset( ( void * ) disk_info, 0, sizeof( mdu_disk_info_t ) ); 

        /* Set the Parent Cookie value in the Disk structure */
        disk_info->parent_cookie = parent_cookie;

        /* set the Index number of this disk */
        disk_info->number = index; 
        
	/* Now, Ask the EMD driver to give us the disk info */
	rc = ioctl(mdctl_fd, EMD_GET_DISK_INFO_FROM_PARENT, disk_info);
//      if  ( rc )   
//             printf( "Error in EMD_GET_DISK_INFO_FROM_PARENT Ioctl = %X : %s\n",
//	       		rc,  strerror(errno));
	return rc;
}


/***************************************************/
/* get_cookie()   Get the cookie for a device name */
/***************************************************/

unsigned long get_cookie(char *device_name)
{
        int rc;
        mdu_cookie_info_t cookie_info;
	unsigned int device = 0;

        /* Zero out the entire local cookie structure */
        memset( ( void * ) &cookie_info, 0, sizeof( mdu_cookie_info_t ) ); 

        /* Set the Device Number ( major, minor ) in the local cookie structure */
        device = get_devnum( device_name );
	cookie_info.major_minor.major = major(device);
	cookie_info.major_minor.minor = minor(device);
        
	rc = ioctl(mdctl_fd, EMD_GET_DEVICE_INFO, &cookie_info); 
//      if  ( rc )      
//            printf( "Error in EMD_GET_DEVICE_INFO Ioctl = %X : %s\n",
//			 rc, strerror(errno));
        
        return( cookie_info.cookie );
}


/***********************************************/
/* get_parent_cookie()   Get EMD Parent Cookie */
/***********************************************/

unsigned long get_parent_cookie( unsigned long cookie)
{
        int rc;
        mdu_cookie_info_t cookie_info;

        /* Zero out the entire local cookie structure */
        memset( ( void * ) &cookie_info, 0, sizeof( mdu_cookie_info_t ) ); 

        /* Set the Cookie in the local cookie structure */
        cookie_info.cookie = cookie;

        rc = ioctl(mdctl_fd, EMD_GET_MEMBER_INFO, &cookie_info); 
//      if  ( rc )      
//              printf( "Error in EMD_GET_MEMBER_INFO Ioctl = %X : %s\n",
//			 rc,  strerror(errno));
        
        return( cookie_info.parent_cookie );
}


/****************************************************/
/* kill_meta_data()   Zero the MetaData on a device */
/****************************************************/

int kill_meta_data( char *devname)
{
        int rc;
        dev_t device;
        mdu_major_minor_t major_minor; 

        device = get_devnum( devname ); 
      	major_minor.major = major(device);
      	major_minor.minor = minor(device);
        /* Ask the EMD driver to Kill the MetaData */
	rc = ioctl(mdctl_fd, EMD_KILL_META_DATA, &major_minor);
        if  ( rc )  
               printf( "Unable to zero super block : %s\n", strerror(errno));
	return rc;
}


/***************************/
/* get_task_status()   TBD */
/***************************/

int get_task_status(char *devname)
{
        int rc;
        dev_t device;

        device = get_devnum( devname ); 

        /* Ask the EMD driver for the Background Task Status */
	rc = ioctl(mdctl_fd, EMD_GET_TASK_STATUS_INFO, device);
//      if  ( rc ) 
//             printf( "Error in EMD_GET_TASK_STATUS Ioctl = %X : %s\n",
//      		 rc,  strerror(errno));

	return rc;
}


/*********************/
/* compare_array()   */
/*********************/

int compare_array(mdu_array_info_t *array1, mdu_array_info_t *array2)
{
       /*
        * return:  
        *          0 same, or first was empty, and second was copied
        *          1  not the same
        */
	
  	if ((array1->major_version == 0) &&
	    (array1->minor_version == 0) &&
    	    (array1->patch_version == 0) &&
       	    (array1->raid_disks == 0)) {
		memcpy(array1, array2, sizeof(*array1));
		return 0;
	}

	if (array1->major_version != array2->major_version ||
	    array1->minor_version != array2->minor_version ||
    	    array1->patch_version != array2->patch_version ||
	    array1->level         != array2->level         ||
	    array1->size          != array2->size          ||
	    array1->nr_disks      != array2->nr_disks      ||
	    array1->raid_disks    != array2->raid_disks    ||
	    array1->md_minor      != array2->md_minor      )
            return 1;
        
	return 0;
}


/***************************************************/
/* get_gencount()   Get the cookie for a device name */
/***************************************************/

unsigned long get_gencount(char *device_name)
{
        int rc;
        mdu_cookie_info_t cookie_info;
	unsigned int device = 0;

        /* Zero out the entire local cookie structure */
        memset( ( void * ) &cookie_info, 0, sizeof( mdu_cookie_info_t ) ); 

        /* Set the Device Number ( major, minor ) in the local cookie structure */
        device = get_devnum( device_name );
	cookie_info.major_minor.major = major(device);
	cookie_info.major_minor.minor = minor(device);
        
	rc = ioctl(mdctl_fd, EMD_GET_DEVICE_INFO, &cookie_info); 
//      if  ( rc )      
//              printf( "Error in EMD_GET_DEVICE_INFO Ioctl = %X : %s\n",
//			 rc,  strerror(errno));
        
        return( cookie_info.gencount );
}


/****************************************************************************/
/* print_array()   For debugging - print the contents of an array structure */
/****************************************************************************/

void print_array(mdu_array_info_t *array)
{
        time_t atime;

        printf( "ARRAY STRUCTURE CONTENTS ( at addr %lX ):\n", ( unsigned long ) array );
        printf( "  Major Version  %X\n", array->major_version );
	printf( "  Minor Version  %X\n", array->minor_version );
	printf( "  Patch version  %X\n", array->patch_version );
        atime = array->utime; 
	printf( "  ctime          %.24s\n", ctime( &atime ) );
	printf( "  Level          %X\n", array->level );
	printf( "  Size           %X\n", array->size );
	printf( "  NR Disks       %X\n", array->nr_disks );
	printf( "  Raid Disks     %X\n", array->raid_disks );
	printf( "  MD Minor       %X\n", array->md_minor );
	printf( "  Not Persistent %X\n", array->not_persistent );
        atime = array->utime; 
	printf( "  utime          %.24s\n", ctime( &atime ) );		
	printf( "  state          %X\n", array->state );		
	printf( "  Active Disks   %X\n", array->active_disks );	
	printf( "  Working Disks  %X\n", array->working_disks );	
	printf( "  Failed Disks   %X\n", array->failed_disks );	
	printf( "  Layout         %X\n", array->layout );		
	printf( "  Chunk Size     %X\n", array->chunk_size );	
        //printf( "  Metadata Type  %X\n", array->metadata_type );
        printf( "  Metadata Type  %s\n", array->metaname);
        printf( "  Parent Cookie  %lX\n", array->parent_cookie );            
        printf( "  Cookie         %lX\n\n", array->cookie );                
}


/*************************************************************************/
/* print_disk()   For debugging - print the contents of a disk structure */
/*************************************************************************/

void print_disk(mdu_disk_info_t *disk)
{
        printf( "DISK STRUCTURE CONTENTS ( at addr %lX ):\n", ( unsigned long ) disk );
        printf( "  Number         %X\n", disk->number );
        printf( "  Major          %X\n", disk->major );
        printf( "  Minor          %X\n", disk->minor );
        printf( "  Raid Disk      %X\n", disk->raid_disk );
        printf( "  State          %X\n", disk->state );
        printf( "  Parent Cookie  %lX\n", disk->parent_cookie );
        printf( "  Cookie         %lX\n", disk->cookie );
        printf( "  Device Type    %X\n\n", disk->device_type );
}

int validate_metapers(int mdctl, char *metaname, int verbose)
{
	int count, i;
	char *namebuf, *name;
	
	if (ioctl(mdctl, EMD_GET_NUM_OF_METAPERS, &count)){
		fprintf(stderr, "Failed to validate metapers %s\n", metaname);
		return -1;
	}
	namebuf = malloc(count * MD_METAPERS_NAME_LENGTH);
	if (!namebuf){
		fprintf(stderr,"No memory\n");
		return -1;
	}
	if (ioctl(mdctl, EMD_GET_METAPERS, namebuf)){
		fprintf(stderr, "Failed to validate metapers %s\n", metaname);
		free(namebuf);
		return -1;
	}
	for (i = 0 ; i < count ; i++){
		name = namebuf + (MD_METAPERS_NAME_LENGTH * i);
		if (strncmp(name, metaname, MD_METAPERS_NAME_LENGTH) == 0){
			free(namebuf);
			return 0;
		}
	}
	if (verbose){
		fprintf(stderr,"Only following metapers are available. "
		"Select anyone among them.\n");
		for (i = 0 ; i < count ; i++){
			name = namebuf + (MD_METAPERS_NAME_LENGTH * i);
			fprintf(stderr, "[%s] ", name); 
		}
		fprintf(stderr, "\n");
	}
	free(namebuf);
	return -1;
}


